#include <stdlib.h>
#include <string.h>

#include <openssl/buffer.h>
#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/err.h>

#include "cryptutil.h"


struct crypthelper_struct
{
    char *ptr;
    long len;
};


static const char k_magic[] = "Salted__";


crypthelper_t* crypthelper_alloc()
{
    crypthelper_t *helper = (crypthelper_t *)malloc(sizeof(crypthelper_t));
    if (helper) {
        helper->ptr = NULL;
        helper->len = 0;
    }
    return helper;
}


void crypthelper_free(crypthelper_t *helper)
{
    if (helper) {
        free(helper->ptr);
        free(helper);
    }
}


const char* crypthelper_getptr(const crypthelper_t *helper)
{
    return helper->ptr;
}


long crypthelper_getlen(const crypthelper_t *helper)
{
    return helper->len;
}


int cryptutil_encrypt(crypthelper_t *helper, const char *in, int in_len,
            const char *pw, int pw_len)
{
    const EVP_CIPHER *cipher;
    const EVP_MD *dgst;
    unsigned char salt[PKCS5_SALT_LEN];
    unsigned char key[EVP_MAX_KEY_LENGTH];
    unsigned char iv[EVP_MAX_IV_LENGTH];
    BIO *bio;
    BIO *bio_enc;
    BUF_MEM *mem;
    int ret = -1;
    
    cipher = EVP_aes_256_cbc();
    dgst = EVP_md5();
    RAND_pseudo_bytes(salt, sizeof salt);
    EVP_BytesToKey(cipher, dgst, salt, (unsigned char*)pw, pw_len, 1, key, iv);
    
    bio = BIO_new(BIO_s_mem());
    bio_enc = BIO_new(BIO_f_cipher());
    if (bio && bio_enc) {
        BIO_set_close(bio, BIO_CLOSE);
        
        BIO_write(bio, k_magic, sizeof k_magic-1);
        BIO_write(bio, salt, sizeof salt);
        
        BIO_set_cipher(bio_enc, cipher, key, iv, 1);
        bio = BIO_push(bio_enc, bio);
        
        BIO_write(bio, in, in_len);
        if (BIO_flush(bio)) {
            BIO_get_mem_ptr(bio, &mem);
            helper->ptr = (char *)malloc(mem->length * sizeof(unsigned char));
            if (helper->ptr) {
                memmove(helper->ptr, mem->data, mem->length);
                helper->len = mem->length;
                ret = 0;
            }
        }
        BIO_free_all(bio);
    } else {
        if (bio) BIO_free(bio);
        if (bio_enc) BIO_free(bio_enc);
    }
    
    return ret;
}


int cryptutil_decrypt(crypthelper_t *helper, const char *in, int in_len,
            const char *pw, int pw_len)
{
    const EVP_CIPHER *cipher;
    const EVP_MD *dgst;
    unsigned char salt[PKCS5_SALT_LEN];
    unsigned char key[EVP_MAX_KEY_LENGTH];
    unsigned char iv[EVP_MAX_IV_LENGTH];
    BIO *bio;
    BIO *bio_enc;
    int ret = -1;
    const char *p = in;
    BUF_MEM *mem;
    
    if (in_len < sizeof k_magic - 1 + sizeof salt) return ret;
    
    if (memcmp(p, k_magic, sizeof k_magic-1)) return ret;
    p += sizeof k_magic - 1;
    memmove(salt, p, sizeof salt);
    p += sizeof salt;
    
    cipher = EVP_aes_256_cbc();
    dgst = EVP_md5();
    EVP_BytesToKey(cipher, dgst, salt, (unsigned char*)pw, pw_len, 1, key, iv);
    
    bio = BIO_new(BIO_s_mem());
    bio_enc = BIO_new(BIO_f_cipher());
    if (bio && bio_enc) {
        BIO_set_cipher(bio_enc, cipher, key, iv, 0);
        bio = BIO_push(bio_enc, bio);
        BIO_write(bio, (unsigned char*)p, in_len - (p - in));
        if (BIO_flush(bio)) {
            BIO_get_mem_ptr(bio, &mem);
            helper->ptr = (char *)malloc(mem->length);
            if (helper->ptr) {
                helper->len = mem->length;
                memmove(helper->ptr, mem->data, mem->length);
                ret = 0;
            }
        }
        BIO_free_all(bio);
    } else {
        if (bio_enc) BIO_free(bio_enc);
        if (bio) BIO_free(bio);
    }
    
    return ret;
}


void cleanup_openssl_library()
{
    ERR_free_strings();
    EVP_cleanup();
    CRYPTO_cleanup_all_ex_data();
    ERR_remove_state(0);
}

